import numpy as np

from hbs.mesh import Mesh
from ..utils.tool_functions import mu_chop, to_complex

def get_beltrami_coefficient(
    mapping: np.ndarray[np.floating], mesh: Mesh
) -> np.ndarray[np.complexfloating]:
    """
    Calculate the Beltrami coefficient of the mapping from "vertex" to "mapping", i.e., f(vertex) = map.
    :param mapping: n x 2 mapped vertex coordinates
    :return: Corresponding Beltrami coefficient
    """
    assert isinstance(mesh, Mesh), "mesh must be Mesh object"

    assert isinstance(mapping, np.ndarray) and np.issubdtype(
        mapping.dtype, np.floating
    ), "mapping must be float array"
    assert mapping.ndim == 2 and mapping.shape == (mesh.vert_num, 2), (
        "mapping must be n x 2 array and n is the number of faces"
    )

    mapping = to_complex(mapping)
    mu = (mesh.Dc * mapping) / (mesh.Dz * mapping)
    mu = mu_chop(mu)
    return mu



import numpy as np

from hbs.conformal_welding import ConformalWelding, get_conformal_welding
from hbs.mesh import DiskMesh, get_unit_disk

from .qc import get_beltrami_coefficient, lsqc_solver
from .utils.geodesic_welding import geodesic_welding
from .utils.poisson import integral as poisson_integral
from .utils.tool_functions import to_complex, to_real


def get_hbs(
    bound: np.ndarray[np.floating],
    circle_point_num: int = 500,
    density: float = 0.01,
    disk: DiskMesh = None,
) -> tuple[
    np.ndarray[np.complexfloating],
    np.ndarray[np.floating],
    ConformalWelding,
    DiskMesh,
]:
    """
    Get Beltrami coefficients from boundary points
    :param `bound`: `n x 2` array, boundary points
    :param `circle_point_num`: number of points on the circle
    :param `density`: density of the mesh
    :param `disk`: DiskMesh object
    :return:
        `hbs`: `n` x 1 complex array, Beltrami coefficients defined on triangles
        `he`: `n` x 1 float array, harmonice extension of `hbs`
        `cw`: ConformalWelding object
        `disk`: DiskMesh object
    """
    if disk is None:
        disk = get_unit_disk(density, circle_point_num)

    cw = get_conformal_welding(bound)

    r = 0
    while True:
        cw.rotate_x(r / 2)
        cw.linear_interp(circle_point_num)
        he = poisson_integral(disk.in_vert, cw.x, cw.y)
        hbs = get_beltrami_coefficient(he, disk)

        excluded_idx = np.isnan(hbs) + np.linalg.norm(disk.face_center, axis=1) == 0
        r = np.angle(np.sum(hbs[~excluded_idx]))

        if abs(r) <= 5e-3:
            he_angle = np.angle(np.sum(hbs * to_complex(disk.face_center)))
            if he_angle < 0 or he_angle == np.pi:
                r += 2 * np.pi
            else:
                break

    return hbs, he, cw, disk


def reconstruct_from_hbs(hbs: np.ndarray[np.complexfloating], disk: DiskMesh):
    """
    Reconstruct original shape from HBS
    :param `hbs`: complex array with length `disk.face_num`, Beltrami coefficients defined on triangles
    :param `disk`: DiskMesh object
    :return:
        `bound_points`: `disk.circle_num` x 2 array, boundary points
        `in_points`: `disk.in_vert_num` x 2 array, inner points
        `out_points`: `disk.out_vert_num` x 2 array, outer points
        `mapping`: `disk.vert_num` x 2 array, vertex coordinates of the solved mapping
    """
    assert isinstance(disk, DiskMesh), "`disk` must be DiskMesh object"

    assert isinstance(hbs, np.ndarray) and np.issubdtype(
        hbs.dtype, np.complexfloating
    ), "`hbs` must be complex array"
    assert hbs.ndim == 1 and hbs.shape[0] == disk.face_num, (
        "the length of `hbs` is `n` and `n` is the number of faces"
    )

    target = np.array([[1, 0], [0, 0]], dtype=np.float64)

    one_pos = np.all(disk.vert == target[0], axis=1)
    zero_pos = np.all(disk.vert == target[1], axis=1)
    landmark = np.where(one_pos | zero_pos)[0]

    mapping = lsqc_solver(hbs, landmark, target, disk)
    bound_points = mapping[: disk.circle_num]
    in_points = mapping[: disk.in_vert_num + disk.circle_num]
    out_points = mapping[disk.in_vert_num + disk.circle_num :]

    in_points, out_points = geodesic_welding(
        to_complex(in_points),
        to_complex(out_points),
        to_complex(bound_points),
        to_complex(disk.circle),
    )
    in_points = to_real(in_points)
    out_points = to_real(out_points)

    bound_points = in_points[: disk.circle_num]
    in_points = in_points[disk.circle_num :]
    return bound_points, in_points, out_points, mapping
    return bound_points, in_points, out_points, mapping



import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d

from .utils.mobius import mobius, mobius_inv
from .utils.tool_functions import to_complex
from .utils.zipper import zipper, zipper_params


class ConformalWelding:
    def __init__(
        self,
        x: np.ndarray[np.complexfloating],
        y: np.ndarray[np.complexfloating],
        params: np.ndarray[np.complexfloating],
    ):
        """
        Initialize the ConformalWelding class.

        Parameters:
        x (np.ndarray): Complex ndarray of shape (n,) representing initial x points.
        y (np.ndarray): Complex ndarray of shape (n,) representing initial y points.
        params (np.ndarray): Complex ndarray of shape (n+1,) representing parameters.
        """
        assert (
            x.ndim == y.ndim == params.ndim == 1
            and np.issubdtype(x.dtype, np.complexfloating)
            and np.issubdtype(y.dtype, np.complexfloating)
            and np.issubdtype(params.dtype, np.complexfloating)
        ), "Input must be 1D complex array"
        assert len(x) == len(y) == len(params) - 1, (
            "Length of x, y, params must be n, n, n+1"
        )

        self.params = params
        self._set_init_x(x)
        self._set_init_y(y)

    def _set_init_x(self, x: np.ndarray[np.complexfloating]):
        x_angle = np.angle(x)
        x_angle_diff = np.diff(x_angle)
        min_pos = np.argmin(x_angle_diff)
        if x_angle_diff[min_pos] < 0:
            x_angle[min_pos + 1 :] += 2 * np.pi
        # x_angle -= x_angle[0]
        # x_angle = np.mod(x_angle, 2 * np.pi)
        x = np.exp(x_angle * 1j)

        self.init_x_angle = x_angle
        self.init_x = x
        self.x = x

    def _set_init_y(self, y: np.ndarray[np.complexfloating]):
        self.init_y = y
        self.y = y

    def uniquify(self):
        x_angle, idx = np.unique(np.angle(self.init_x), return_index=True)
        x = np.exp(x_angle * 1j)
        y = self.init_y[idx]
        y = np.exp(np.angle(y) * 1j)

        self._set_init_x(x)
        self._set_init_y(y)

    def linear_interp(self, num: int):
        x_angle_regular = np.arange(0, 2 * np.pi, 2 * np.pi / num)
        self.x = np.exp(x_angle_regular * 1j)

        inter_func = interp1d(
            np.concatenate(
                [
                    self.init_x_angle - 2 * np.pi,
                    self.init_x_angle,
                    self.init_x_angle + 2 * np.pi,
                ]
            ),
            np.concatenate([self.init_y, self.init_y, self.init_y]),
            kind="linear",
        )
        y = inter_func(x_angle_regular)
        y_anlge = np.angle(y)
        self.y = np.exp((y_anlge - y_anlge[0]) * 1j)

    def rotate_x(self, r):
        x_angle = np.mod(self.init_x_angle - r, 2 * np.pi)
        x = np.exp(x_angle * 1j)
        self._set_init_x(x)

    def x_post_norm(self):
        """
        对外部点进行后归一化
        :param z: 输入点
        :param params: 参数
        :return: 归一化后的点，a 和 theta 参数
        """
        zinf = zipper_params([np.inf], self.params)
        a = -1 / np.conj(zinf)
        theta = 0
        x = mobius_inv(self.x, a, theta)
        x = np.flipud(x)
        self._set_init_x(x)

    def y_post_norm(self):
        """
        对内部点进行后归一化
        :param z: 输入点
        :return: 归一化后的点，a 和 theta 参数
        """
        p = np.append(self.y, [0, 1])
        while True:
            center = np.mean(p[:-2])
            if np.abs(center) <= np.finfo(float).eps:
                break
            p = mobius(p, center)

        y = p[:-2]
        # y_angle = self.y_angle_norm()
        # y = np.exp(y_angle * 1j)
        self._set_init_y(y)

    def plot_x(self):
        plt.gca().set_aspect("equal", adjustable="box")
        plt.scatter(self.x.real, self.x.imag)
        plt.show()

    def plot_y(self):
        plt.gca().set_aspect("equal", adjustable="box")
        plt.scatter(self.y.real, self.y.imag)
        plt.show()

    def plot(self, is_interp=True):
        plt.gca().set_aspect("equal", adjustable="box")
        x_angle = np.angle(self.x)
        x_angle = np.mod(x_angle, 2 * np.pi)
        y_angle = self.y_angle_norm()

        if is_interp:
            plt.plot(x_angle, y_angle, linestyle="-", linewidth=2)
        else:
            plt.scatter(x_angle, y_angle, s=2)
        plt.show()

    def y_angle_norm(self):
        y_angle = np.angle(self.y)
        y_angle_diff = np.diff(y_angle)

        # plt.plot(y_angle)
        # plt.plot(y_angle_diff)
        # plt.show()

        y_angle_diff = np.insert(y_angle_diff, 0, 0)
        y_angle_diff[y_angle_diff < -np.pi] += 2 * np.pi
        y_angle_diff[(-0.1 < y_angle_diff) & (y_angle_diff < 0)] *= -1

        if np.any(y_angle_diff < 0):
            raise ValueError("y_angle_diff must be non-negative")

        if y_angle_diff.sum() > 2 * np.pi:
            y_angle_diff *= 2 * np.pi / (y_angle_diff.sum() + y_angle_diff.mean())

        y_angle = np.cumsum(y_angle_diff)
        return y_angle


def get_conformal_welding(bound: np.ndarray[np.floating]) -> ConformalWelding:
    assert (
        bound.ndim == 2
        and bound.shape[1] == 2
        and np.issubdtype(bound.dtype, np.floating)
    ), "bound must be n x 2 real array with float type"

    bound = to_complex(bound)
    x, _, x_params = zipper(bound)
    # x = np.flipud(x)
    y, _, _ = zipper(np.flipud(bound))

    cw = ConformalWelding(x, y, x_params)
    cw.x_post_norm()
    cw.y_post_norm()
    cw.uniquify()
    return cw



import numpy as np
from scipy.sparse import csr_matrix, lil_matrix
from scipy.sparse.linalg import spsolve

from hbs.mesh import Mesh


def lsqc_solver(
    mu: np.ndarray[np.complexfloating],
    landmark: np.ndarray[np.integer],
    target: np.ndarray[np.floating],
    mesh: Mesh
):
    """
    Solves the Beltrami equation
    :param mu: n x 1 complex array, Beltrami coefficients defined on triangles
    :param landmark: m x 1 array, indices of landmark vertices
    :param target: m x 2 array, target positions of constrained vertices
    :param mesh: Mesh object
    :return: vertex coordinates of the solved mesh
    """
    assert isinstance(mesh, Mesh), "mesh must be Mesh object"

    assert isinstance(mu, np.ndarray) and np.issubdtype(mu.dtype, np.complexfloating), (
        "mu must be complex array"
    )
    assert mu.ndim == 1 and mu.shape[0] == mesh.face_num, (
        "mu must be n x 1 array and n is the number of faces"
    )

    assert isinstance(landmark, np.ndarray) and np.issubdtype(
        landmark.dtype, np.integer
    ), "landmark must be integer array"
    assert landmark.ndim == 1, "landmark must be 1D array"

    assert isinstance(target, np.ndarray) and np.issubdtype(
        target.dtype, np.floating
    ), "target must be float array"
    assert (
        target.ndim == 2
        and target.shape == (landmark.shape[0], 2)
    ), "target must be m x 2 array and m is the number of landmarks"

    landmark = np.concatenate([landmark, landmark + mesh.vert_num])
    target = target.T.reshape(-1)
    L = div_PtP_grad(mu, mesh)
    A = unsigned_area_matrix(mu, mesh)
    M = lil_matrix(L - 2 * A)
    b = -M[:, landmark] @ target
    M[landmark, :] = 0
    M[:, landmark] = 0
    M[landmark, landmark] = 1
    M = M.tocsr()
    u = spsolve(M, b)
    u[landmark] = target
    u = u.reshape(2, -1).T
    return u


def unsigned_area_matrix(mu, mesh):
    idx = np.abs(mu) > 1
    A = np.ones(len(mu))
    A[idx] = -A[idx]

    def compute_q(ei, ej):
        yi, xi = ei.T
        yj, xj = ej.T
        return A * (yi * xj - xi * yj) / (4 * mesh.area)

    x_order = [0, 1, 2, 0, 1, 0, 2, 1, 2]
    y_order = [0, 1, 2, 1, 0, 2, 0, 2, 1]

    x_pos = np.concatenate([mesh.face[:, k] for k in x_order])
    y_pos = np.concatenate([mesh.face[:, k] for k in y_order])
    value = np.concatenate(
        [compute_q(mesh.edges[i], mesh.edges[j]) for i, j in zip(x_order, y_order)]
    )

    x_pos = np.concatenate([x_pos, x_pos + mesh.vert_num])
    y_pos = np.concatenate([y_pos + mesh.vert_num, y_pos])
    value = np.concatenate([value, -value]) / 2
    matrix = csr_matrix((value, (x_pos, y_pos)))
    return matrix


def div_PtP_grad(mu, mesh):
    abs_mu = np.abs(mu)
    A = -(1 - 2 * np.real(mu) + abs_mu**2) / (1.0 - abs_mu**2)
    B = 2 * np.imag(mu) / (1.0 - abs_mu**2)
    C = -(1 + 2 * np.real(mu) + abs_mu**2) / (1.0 - abs_mu**2)

    idx = abs_mu > 1
    A[idx] = -A[idx]
    B[idx] = -B[idx]
    C[idx] = -C[idx]

    def compute_v(ei, ej):
        yi, xi = ei.T
        yj, xj = ej.T
        return (A * (xi * xj) - B * (xi * yj + xj * yi) + C * (yi * yj)) / mesh.area

    x_order = [0, 1, 2, 0, 1, 1, 2, 2, 0]
    y_order = [0, 1, 2, 1, 0, 2, 1, 0, 2]

    x_pos = np.concatenate([mesh.face[:, k] for k in x_order])
    y_pos = np.concatenate([mesh.face[:, k] for k in y_order])
    value = np.concatenate(
        [compute_v(mesh.edges[i], mesh.edges[j]) for i, j in zip(x_order, y_order)]
    )

    x_pos = np.concatenate([x_pos, x_pos + mesh.vert_num])
    y_pos = np.concatenate([y_pos, y_pos + mesh.vert_num])
    value = -np.concatenate([value, value]) / 4
    matrix = csr_matrix((value, (x_pos, y_pos)))
    return matrix



import numpy as np


def mu_chop(mu, bound=0.9999, constant=None):
    """
    Perform truncation on mu
    :param mu: Input complex array
    :param bound: Truncation boundary value
    :param constant: Constant value after truncation
    :return: Processed mu
    """
    if constant is None:
        constant = bound

    abs_mu = np.abs(mu)
    idx = abs_mu >= bound
    mu[idx] = constant * (mu[idx] / abs_mu[idx])
    return mu


def to_real(x: np.ndarray[np.complexfloating]) -> np.ndarray[np.floating]:
    assert isinstance(x, np.ndarray) and np.issubdtype(x.dtype, np.complexfloating), (
        "x must be a complex array"
    )
    assert x.ndim == 1, "x array must be 1D array"
    return np.stack([x.real, x.imag], axis=-1)


def to_complex(x: np.ndarray[np.floating]) -> np.ndarray[np.complexfloating]:
    # assert , "Input must be a numpy array"
    # assert np.issubdtype(x.dtype, np.floating) or np.issubdtype(x.dtype, np.integer), (
    #     "Input array must have floating point or integer data type"
    # )
    assert isinstance(x, np.ndarray) and np.issubdtype(x.dtype, np.floating), (
        "x must be a real array"
    )
    assert x.ndim == 2 and x.shape[1] == 2, "x must be n x 2 array"
    return x[:, 0] + 1j * x[:, 1]



import cv2
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import CubicSpline, PchipInterpolator


def get_boundary(image_path, num_points=250):
    """
    Process the image to extract boundary, apply anti-aliasing, and resample.

    Parameters:
    image_path: Input image path
    num_points: Number of resampled points

    Returns:
    resampled_points: List of resampled boundary points
    """
    # Extract boundary points
    boundary_points = extract_boundary_points(image_path)
    # Apply anti-aliasing
    smoothed_points = anti_aliasing(boundary_points)
    # Resample to the specified number of points
    resampled_points = smooth_resample(smoothed_points, num_points)
    return resampled_points


def extract_boundary_points(image_path):
    """
    Extract ordered boundary points from a near black-and-white image.

    Parameters:
    image_path: Input image path

    Returns:
    boundary_points: List of ordered boundary points, format [(x1, y1), (x2, y2), ...]
    """
    # Read the image and convert to grayscale
    img = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)

    # Add padding to avoid detecting image borders
    img = cv2.copyMakeBorder(img, 1, 1, 1, 1, cv2.BORDER_CONSTANT, value=255)

    # Determine if the image is black-on-white or white-on-black
    # Calculate the mean color of the outer two layers of pixels
    outer_layer_mean = np.mean(
        [img[:2, :].mean(), img[-2:, :].mean(), img[:, :2].mean(), img[:, -2:].mean()]
    )

    if outer_layer_mean > 127:
        # White-on-black image, invert it
        img = cv2.bitwise_not(img)

    # Binarize the image
    _, binary = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY)

    # Find contours (using CHAIN_APPROX_NONE to get all boundary points)
    contours, _ = cv2.findContours(binary, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)

    # Get the largest contour (assuming there's only one main shape in the image)
    main_contour = max(contours, key=cv2.contourArea)

    # Return the contour points directly, without sorting or anti-aliasing
    return main_contour.reshape(-1, 2)


def anti_aliasing(points, kernel_size=5):
    """
    Apply anti-aliasing to boundary points.

    Parameters:
    points: List of original boundary points

    Returns:
    smoothed_points: List of smoothed boundary points
    """
    # Use Gaussian blur for smoothing
    x = cv2.GaussianBlur(points[:, 0].astype("float32"), (kernel_size, kernel_size), 0)
    y = cv2.GaussianBlur(points[:, 1].astype("float32"), (kernel_size, kernel_size), 0)
    return np.column_stack((x, y))


def plot_boundary(boundary, with_line=True, scale=1):
    num_points = boundary.shape[0]
    x_min, x_max = np.min(boundary[:, 0]), np.max(boundary[:, 0])
    y_min, y_max = np.min(boundary[:, 1]), np.max(boundary[:, 1])
    aspect_ratio = (x_max - x_min) / (y_max - y_min)

    fig_size = (num_points / 60 * scale * aspect_ratio, num_points / 60 * scale)
    plt.figure(figsize=fig_size)
    if with_line:
        plt.plot(
            boundary[:, 0],
            boundary[:, 1],
            linestyle="-",
            linewidth=1 * scale,
            color="white",
            label="Boundary Line",
        )
    plt.scatter(
        boundary[:, 0],
        boundary[:, 1],
        marker="o",
        s=15 * scale,
        edgecolor="lightgreen",
        facecolor="none",
        label="Boundary Points",
    )
    plt.show()


def smooth_resample(points, num_points=None):
    """
    Smoothly and uniformly resample the boundary points matrix, generating a matrix of m points.
    :param points: nx2 matrix, original boundary point coordinates
    :param num_points: Number of resampled points
    :return: mx2 matrix, resampled boundary points
    """
    if num_points is None:
        num_points = points.shape[0]

    # Check if the boundary is closed
    is_closed = np.all(np.abs(points[0, :] - points[-1, :]) < 1e-8)
    if not is_closed:
        points = np.vstack([points, points[0, :]])  # Close the boundary if not closed

    # Calculate cumulative arc length
    diffs = np.diff(points, axis=0)
    dists = np.sqrt(np.sum(diffs**2, axis=1))
    cum_dists = np.concatenate([[0], np.cumsum(dists)])
    total_length = cum_dists[-1]

    # Spline interpolation settings
    t = cum_dists
    x = points[:, 0]
    y = points[:, 1]

    if is_closed:
        # Periodic cubic spline interpolation (closed curve)
        ppx = CubicSpline(t, x, bc_type="periodic")
        ppy = CubicSpline(t, y, bc_type="periodic")
    else:
        # Shape-preserving piecewise cubic interpolation (open curve)
        ppx = PchipInterpolator(t, x)
        ppy = PchipInterpolator(t, y)

    # Generate uniform parameters
    if is_closed:
        t_new = np.linspace(0, total_length, num_points + 2)
        t_new = t_new[:-1]  # Avoid duplicate start and end points
    else:
        t_new = np.linspace(0, total_length, num_points + 1)

    # Calculate new points
    x_new = ppx(t_new)
    y_new = ppy(t_new)
    resampled_points = np.column_stack([x_new, y_new])
    return resampled_points[:-1, :]


if __name__ == "__main__":
    # 示例用法
    boundary_points = extract_boundary_points("test_image.png")
    print(f"提取的边界点数量：{len(boundary_points)}")



import numpy as np

def mobius(z, a, theta=None):
    if theta is None:
        k = 1
    else:
        k = np.exp(theta * 1j)

    if a != 0:
        w = k * (z - a) / (1 - np.conj(a) * z)
        w[np.isinf(z)] = -k / np.conj(a)
        w[np.isclose(z, 1 / np.conj(a), atol=1e-15)] = np.inf
    else:
        w = k * z
    return w


def mobius_inv(z, a, theta=None):
    if theta is None:
        return mobius(z, -a)
    else:
        k = np.exp(theta * 1j)
    return mobius(z, -k * a, -theta)


def mobius_d(z, a, theta):
    return np.exp(theta * 1j) * (1 - abs(a) ** 2) / (1 - np.conj(a) * z) ** 2


import numpy as np


def zipper(bound, others=None):
    if others is None:
        others = []

    n = len(bound)
    params = np.zeros(n + 1, dtype=complex)
    params[:2] = bound[:2]
    points = f_pre(np.concatenate([bound, others]), params[0], params[1])
    for j in range(2, n):
        params[j] = points[j]
        points = f(points, points[j])
    params[n] = points[0]
    points = f_end(points, params[n])
    points = f_final(points)

    bound = points[:n]
    others = points[n:]
    return bound, others, params



def f_pre(z, p, q):
    """
    map z to right half plane, where real part is positive.
    p -> inf,
    q -> 0,
    note that always p = z[0], q = z[1]
    """
    with np.errstate(divide="ignore", invalid="ignore"):
        w = (z - q) / (z - p)
    w[np.isinf(z)] = 1
    w[z == p] = np.inf
    w = np.sqrt(w)
    return w


def f1(z, p):
    """
    0 -> 0
    p -> 1
    aj -> bj
    inf -> - c/d * 1j
    1j/d -> inf
    """
    c = np.real(p) / np.abs(p) ** 2
    d = np.imag(p) / np.abs(p) ** 2
    # if d == 0:
    #     s = z[:10]
    #     print(s, p)
    #     print(s.imag)
    #     from matplotlib import pyplot as plt
    #     plt.scatter(s.real, s.imag)
    #     plt.show()
    if d == 0:
        with np.errstate(divide="ignore", invalid="ignore"):
            w = z * c
        w[np.isinf(z)] = np.inf
    else:
        with np.errstate(divide="ignore", invalid="ignore"):
            w = c * z / (1  + 1j * d * z)
        w[np.isinf(z)] = -c / d * 1j
        w[z == 1j / d] = np.inf
        w[np.isclose(z, p)] = 1
    return w


def f2(z):
    """
    1 -> 0
    0 -> 1j
    bj -> sqrt(b^2 + 1)j (positive imaginary part)
    """
    with np.errstate(divide="ignore", invalid="ignore"):
        w = np.sqrt(z**2 - 1)
    k = np.imag(w) * np.imag(z) < 0
    w[k] = -w[k]
    w[z == 0] = 1j
    w[np.isinf(z)] = np.inf
    return w


def f(z, p):
    """
    map p to 0,
    with keep points in y-axis, going up.
    0 -> 0 -> 1j
    p -> 1 -> 0
    aj -> bj -> sqrt(b^2 + 1)j
    """
    if p.real == 0:
        raise ValueError("Point p should not be a pure imaginary number.")

    w = f1(z, p)
    w = f2(w)
    w[np.isclose(z, p)] = 0
    return w


def f_end(z, p):
    q = 1 - z / p
    with np.errstate(divide="ignore", invalid="ignore"):
        w = (z / q) ** 2
    w[np.isinf(z)] = p**2
    w[q == 0] = np.inf
    return w


def f_final(z):
    with np.errstate(divide="ignore", invalid="ignore"):
        w = (z - 1j) / (z + 1j)
    w[np.isinf(z)] = 1
    w[z == -1j] = np.inf
    return w


def zipper_params(points, params):
    n = len(params) - 1
    points = f_pre(points, params[0], params[1])
    for j in range(2, n):
        points = f(points, params[j])
    points = f_end(points, params[n])
    points = f_final(points)
    return points


def zipper_inv(points, params):
    n = len(params) - 1
    points = f_final_inv(points)
    points = f_end_inv(points, params[n])
    for j in range(n - 1, 1, -1):
        points = f_inv(points, params[j])
    points = f_pre_inv(points, params[0], params[1])
    return points


def f_pre_inv(w, p, q):
    z = (p * w**2 - q) / (w**2 - 1)
    z[np.isinf(w)] = p
    z[w == 1] = np.inf
    return z


def f1_inv(w, p):
    pc = np.real(p) / np.abs(p) ** 2
    pd = np.imag(p) / np.abs(p) ** 2
    z = w / (pc - 1j * pd * w)
    z[np.isinf(w)] = 1j / pd
    z[w == -1j * pc / pd] = np.inf
    return z


def f2_inv(w):
    z = np.sqrt(w**2 + 1)
    k = np.imag(w) * np.imag(z) < 0
    z[k] = -z[k]
    z[w == 1j] = 0
    z[np.isinf(w)] = np.inf
    return z


def f_inv(w, p):
    z = f2_inv(w)
    z = f1_inv(z, p)
    return z


def f_end_inv(w, p):
    z = np.sqrt(w)
    with np.errstate(divide="ignore", invalid="ignore"):
        z = z / (1 + z / p)
    z[np.isinf(w)] = p
    return z


def f_final_inv(w):
    with np.errstate(divide="ignore", invalid="ignore"):
        z = (w + 1) * 1j / (1 - w)
    k = np.abs(np.imag(z)) < 1e-10
    z[k] = np.real(z[k])
    z[np.isinf(w)] = -1j
    z[w == 1] = np.inf
    return z


# def zipper_d(points, params):
#     n = len(params) - 1
#     d = f_pre_d(points, params[0], params[1])
#     points = f_pre(points, params[0], params[1])
#     for j in range(2, n):
#         d *= f_d(points, params[j])
#         points = f(points, params[j])
#     d *= f_end_d(points, params[n])
#     points = f_end(points, params[n])
#     d *= f_final_d(points)
#     return d

# def f_pre_d(z, p, q):
#     t = f_pre(z, p, q)
#     d = (q - p) / (2 * ((z - p) ** 2) * t)
#     d[np.isinf(z)] = 0
#     d[z == p] = np.inf
#     d[z == q] = np.inf
#     return d


# def f1_d(z, p):
#     pc = np.real(p) / np.abs(p) ** 2
#     pd = np.imag(p) / np.abs(p) ** 2
#     w = pc / (1 + 1j * pd * z) ** 2
#     w[np.isinf(z)] = 0
#     w[z == 1j / pd] = np.inf
#     return w


# def f2_d(z):
#     d = z / f2(z)
#     d[np.isinf(z)] = 1
#     return d


# def f_d(z, p):
#     t = f1(z, p)
#     d1 = f1_d(z, p)
#     d2 = f2_d(t)
#     d = d1 * d2
#     return d


# def f_end_d(z, p):
#     d = 2 * z / (1 - z / p) ** 3
#     d[np.isinf(z)] = 0
#     d[z == p] = np.inf
#     return d


# def f_final_d(z):
#     d = 2j / (z + 1j) ** 2
#     d[np.isinf(z)] = 0
#     d[z == -1j] = np.inf
#     return d

# def zipper_inv_d(points, params):
#     n = len(params) - 1
#     dl = np.zeros(n + 1, dtype=complex)
#     d = f_final_inv_d(points)
#     dl[n] = np.angle(d)
#     points = f_final_inv(points)
#     d0 = f_end_inv_d(points, params[n])
#     d *= d0
#     dl[n - 1] = np.angle(d)
#     points = f_end_inv(points, params[n])
#     for j in range(n - 1, 1, -1):
#         d0 = f_inv_d(points, params[j])
#         d *= d0
#         dl[j - 1] = np.angle(d)
#         points = f_inv(points, params[j])
#     d0 = f_pre_inv_d(points, params[0], params[1])
#     dl[0] = np.angle(d)
#     d *= d0
#     return d

# def f_pre_inv_d(w, p, q):
#     d = (q - p) * 2 * w / (w**2 - 1) ** 2
#     d[w**2 == 1] = np.inf
#     d[np.isinf(w)] = 0
#     return d


# def f1_inv_d(w, p):
#     pc = np.real(p) / np.abs(p) ** 2
#     pd = np.imag(p) / np.abs(p) ** 2
#     d = pc / (pc - 1j * pd * w) ** 2
#     d[np.isinf(w)] = 0
#     d[w == 1j * pc / pd] = np.inf
#     return d


# def f2_inv_d(w):
#     d = w / f2_inv(w)
#     d[np.isinf(w)] = 1
#     return d


# def f_inv_d(w, p):
#     t = f2_inv(w)
#     d1 = f1_inv_d(t, p)
#     d2 = f2_inv_d(w)
#     d = d1 * d2
#     return d


# def f_end_inv_d(w, p):
#     q = np.sqrt(w)
#     d = 1 / (2 * q * (1 + q / p) ** 2)
#     d[np.isinf(w)] = 0
#     d[q == 0] = np.inf
#     d[q == -p] = np.inf
#     return d


# def f_final_inv_d(w):
#     d = 2j / (1 - w) ** 2
#     d[np.isinf(w)] = -2j
#     d[w == 1] = np.inf
#     return d



import numpy as np

from .tool_functions import to_complex, to_real

def kernel(z, x):
    """
    Compute the Poisson kernel
    :param z: Input points
    :param x: Boundary points
    :return: Poisson kernel
    """
    w = z.reshape(-1, 1) / x.reshape(1, -1)
    w = np.real((1 + w) / (1 - w))
    return w


def integral(
    z: np.ndarray[np.floating],
    x: np.ndarray[np.complexfloating],
    y: np.ndarray[np.complexfloating],
) -> np.ndarray[np.floating]:
    """
    Compute the Poisson integral
    :param z: Internal points to be integrated, n x 2 real array
    :param x: Boundary points, m x 1 complex array
    :param y: Boundary values, m x 1 complex array
    :return: Poisson integral result
    """
    assert z.ndim == 2 and z.shape[1] == 2 and np.issubdtype(z.dtype, np.floating), (
        "z must be n x 2 array with float type"
    )

    assert (
        x.ndim == y.ndim == 1
        and x.shape == y.shape
        and np.issubdtype(x.dtype, np.complexfloating)
        and np.issubdtype(y.dtype, np.complexfloating)
    ), "x and y must be m x 1 array with complex type"

    z = to_complex(z)
    theta = np.angle(x)
    dth = np.diff(theta)
    dth = np.append(dth, theta[0] - theta[-1])
    dth[dth < 0] += 2 * np.pi
    kernel_values = kernel(z, x)
    w = (kernel_values @ (y * dth)) / (2 * np.pi)
    w = np.concatenate([y, w])
    w = to_real(w)
    return w



from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import csr_matrix
from scipy.spatial import Delaunay


class Mesh:
    def __init__(self, vert: np.ndarray, face: np.ndarray | None = None):
        assert vert.ndim == 2
        assert vert.shape[1] == 2
        self.vert = vert

        if face is None:
            self.face = Delaunay(vert).simplices

        else:
            assert face.shape[1] == 3
            assert face.ndim == 2
            self.face = face

        self.vert_num = self.vert.shape[0]
        self.face_num = self.face.shape[0]

        self.face_center = self.get_face_center()
        self.edges = self.get_edge()
        self.area = self.get_area()
        self.set_operator()

    def get_face_center(self):
        return np.mean(self.vert[self.face], axis=1)

    def get_edge(self):
        e1 = self.vert[self.face[:, 2]] - self.vert[self.face[:, 1]]
        e2 = self.vert[self.face[:, 0]] - self.vert[self.face[:, 2]]
        e3 = self.vert[self.face[:, 1]] - self.vert[self.face[:, 0]]
        return e1, e2, e3

    def get_area(self):
        return np.cross(self.edges[0], self.edges[1]) / 2

    def set_operator(self):
        def get_diff():
            Mi = np.repeat(np.arange(self.face_num), 3)
            Mj = self.face.flatten()

            stack_area = np.vstack([self.area, self.area, self.area])

            M = np.stack(self.edges, axis=0)
            Mx = M[:, :, 1] / (2 * stack_area)
            My = -M[:, :, 0] / (2 * stack_area)
            Mx = Mx.T.flatten()
            My = My.T.flatten()

            Dx = csr_matrix((Mx, (Mi, Mj)))
            Dy = csr_matrix((My, (Mi, Mj)))
            Dz = (Dx - 1j * Dy) / 2
            Dc = (Dx + 1j * Dy) / 2
            return Dx, Dy, Dz, Dc

        def f2v():
            ring = defaultdict(list)
            for i, face in enumerate(self.face):
                for vertex in face:
                    ring[vertex].append(i)
            ring = [ring[i] for i in range(self.vert_num)]

            avg = np.array([len(r) for r in ring])

            A = csr_matrix(
                (1 / avg, (np.arange(self.vert_num), np.arange(self.vert_num)))
            )

            x_pos = np.concatenate([np.full(len(r), i) for i, r in enumerate(ring)])
            y_pos = np.concatenate(ring)
            B = csr_matrix(
                (np.ones(len(y_pos)), (x_pos, y_pos)),
                shape=(self.vert_num, self.face_num),
            )
            matrix = A @ B
            return matrix

        def v2f():
            x_pos = np.repeat(np.arange(self.face_num), 3)
            y_pos = self.face.flatten()
            matrix = csr_matrix(
                (np.ones(len(y_pos)) / 3, (x_pos, y_pos)),
                shape=(self.face_num, self.vert_num),
            )
            return matrix

        def laplacian():
            f1, f2, f3 = self.face.T
            e1, e2, e3 = self.edges

            l1 = (e1**2).sum(axis=1)
            l2 = (e2**2).sum(axis=1)
            l3 = (e3**2).sum(axis=1)

            cot12 = (l1 + l2 - l3) / (2 * self.area)
            cot23 = (l2 + l3 - l1) / (2 * self.area)
            cot31 = (l1 + l3 - l2) / (2 * self.area)
            diag1 = -cot12 - cot31
            diag2 = -cot12 - cot23
            diag3 = -cot31 - cot23

            x_pos = np.concatenate([f1, f2, f2, f3, f3, f1, f1, f2, f3])
            y_pos = np.concatenate([f2, f1, f3, f2, f1, f3, f1, f2, f3])
            value = np.concatenate(
                [cot12, cot12, cot23, cot23, cot31, cot31, diag1, diag2, diag3]
            )
            matrix = csr_matrix(
                (value, (x_pos, y_pos)), shape=(self.vert_num, self.vert_num)
            )
            return matrix

        self.Dx, self.Dy, self.Dz, self.Dc = get_diff()
        self.f2v = f2v()
        self.v2f = v2f()
        self.laplacian = laplacian()

    def plot(self, with_face_center: bool = False):
        # plt.figure(figsize=(8, 8))
        plt.gca().set_aspect("equal", adjustable="box")
        plt.triplot(self.vert[:, 0], self.vert[:, 1], self.face)
        if with_face_center:
            plt.plot(self.face_center[:, 0], self.face_center[:, 1], "o")
        plt.show()

    def plot_mu(self, mu: np.ndarray[np.complexfloating], is_3d: bool = False):
        assert isinstance(mu, np.ndarray) and np.issubdtype(
            mu.dtype, np.complexfloating
        ), "mu must be complex array"
        assert mu.ndim == 1 and mu.shape[0] == self.face_num, (
            "mu must be 1D array with length equal to face number"
        )

        magnitude = np.abs(mu)
        magnitude[magnitude > 1] = 1
        # 绘制散点图
        if not is_3d:
            plt.scatter(
                self.face_center[:, 0],
                self.face_center[:, 1],
                c=magnitude,
                cmap="jet",
                s=0.5,
            )
            plt.colorbar(label="Magnitude")
        else:
            ax = plt.axes(projection="3d")
            ax.scatter(
                self.face_center[:, 0],
                self.face_center[:, 1],
                magnitude,
                c=magnitude,
                cmap="jet",
                s=0.5,
            )
        plt.show()

    def plot_mapping(self, mapping: np.ndarray[np.floating], is_3d: bool = False):
        assert isinstance(mapping, np.ndarray) and np.issubdtype(
            mapping.dtype, np.floating
        ), "mapping must be real float array"
        assert mapping.ndim == 2 and mapping.shape == (self.vert_num, 2), (
            "mapping must be n x 2 array"
        )

        magnitude = np.linalg.norm(mapping, axis=1)
        # magnitude[magnitude > 1] = 1
        if not is_3d:
            plt.scatter(
                self.vert[:, 0],
                self.vert[:, 1],
                c=magnitude,
                cmap="jet",
                s=0.5,
            )
            plt.colorbar(label="Magnitude")
        else:
            ax = plt.axes(projection="3d")
            ax.scatter(
                self.vert[:, 0],
                self.vert[:, 1],
                magnitude,
                c=magnitude,
                cmap="jet",
                s=0.5,
            )
        plt.show()


class DiskMesh(Mesh):
    def __init__(
        self,
        circle: np.ndarray,
        in_vert: np.ndarray | None = None,
        out_vert: np.ndarray | None = None,
        face: np.ndarray | None = None,
    ):
        assert circle.ndim == 2
        assert circle.shape[1] == 2
        self.circle = circle
        self.circle_num = self.circle.shape[0]
        vert = circle

        if in_vert is not None:
            assert in_vert.ndim == 2
            assert in_vert.shape[1] == 2

            self.in_vert = in_vert
            self.in_vert_num = self.in_vert.shape[0]
            vert = np.vstack([vert, in_vert])
        else:
            self.in_vert = None
            self.in_vert_num = 0

        if out_vert is not None:
            assert out_vert.ndim == 2
            assert out_vert.shape[1] == 2

            self.out_vert = out_vert
            self.out_vert_num = self.out_vert.shape[0]
            vert = np.vstack([vert, out_vert])
        else:
            self.out_vert = None
            self.out_vert_num = 0

        super().__init__(vert, face)

    def update_out_vert(self, out_vert: np.ndarray):
        assert self.out_vert is None
        assert out_vert.ndim == 2
        assert out_vert.shape[1] == 2

        self.out_vert = out_vert
        self.out_vert_num = self.out_vert.shape[0]
        self.vert = np.vstack([self.vert, self.out_vert])
        self.vert_num = self.vert.shape[0]

        ori_face_num = self.face.shape[0]
        out_face = Delaunay(np.vstack([self.circle, out_vert])).simplices
        out_face[out_face >= self.circle_num] += self.in_vert_num

        self.face = np.vstack([self.face, out_face])
        self.face_center = self.get_face_center()

        out_face_num = out_face.shape[0]
        idx = (
            np.where(np.linalg.norm(self.face_center[-out_face_num:], axis=1) <= 1)[0]
            + ori_face_num
        )

        self.face = np.delete(self.face, idx, axis=0)
        self.face_num = self.face.shape[0]
        self.face_center = np.delete(self.face_center, idx, axis=0)

        self.edges = self.get_edge()
        self.area = self.get_area()
        self.set_operator()


def get_rect(
    height: float | int = 1, width: float | int = 1, density: float | int = 0.01
) -> Mesh:
    """
    Create a rectangle mesh
    :param height: height of the rectangle
    :param width: width of the rectangle
    :param density: density of the mesh
    :return: face and vert arrays
    """
    x_num = int(width / density) + 1
    y_num = int(height / density) + 1
    x, y = np.meshgrid(np.arange(x_num), np.arange(y_num))
    x = x * density
    y = y * density
    vert = np.vstack([x.ravel(), y.ravel()]).T
    mesh = Mesh(vert)
    return mesh


def get_unit_disk(
    density: float | int = 0.01, circle_point_num: int = 1000, eps: float | None = None
) -> DiskMesh:
    """
    Create a unit disk mesh
    :param density: density of the mesh
    :param circle_point_num: number of points on the circle
    :param eps: epsilon to shrink the disk
    :return: face and vert arrays
    """
    if eps is None:
        eps = 10 / circle_point_num

    num = int(1 / density)
    x, y = np.meshgrid(np.arange(-num, num + 1), np.arange(-num, num + 1))
    vert = np.vstack([x.ravel(), y.ravel()]).T * density
    index = np.linalg.norm(vert, axis=1) <= 1 - eps
    vert = vert[index, :]

    if circle_point_num:
        circle = np.array(
            [
                [np.cos(t), np.sin(t)]
                for t in np.linspace(0, 2 * np.pi, circle_point_num + 1)
            ]
        )[:-1]

    mesh = DiskMesh(circle, vert)
    return mesh


def get_unit_disk_in_rect(
    disk_mesh: DiskMesh,
    height: float | int = 4,
    width: float | int = 4,
    density: float | int = 0.01,
    eps: float = 1e-3,
) -> DiskMesh:
    """
    Create a rectangle mesh from a unit disk mesh
    :param disk_mesh: unit disk mesh
    :param height: height of the rectangle
    :param width: width of the rectangle
    :param density: density of the mesh
    :return: combined mesh
    """
    x_num = int(width / density) + 1
    y_num = int(height / density) + 1
    x, y = np.meshgrid(np.arange(x_num), np.arange(y_num))
    x = x * density - width / 2
    y = y * density - height / 2
    out_vert = np.vstack([x.ravel(), y.ravel()]).T
    out_vert = out_vert[np.linalg.norm(out_vert, axis=1) > 1 + eps, :]

    new_disk_mesh = DiskMesh(disk_mesh.circle, disk_mesh.in_vert)
    new_disk_mesh.update_out_vert(out_vert)
    return new_disk_mesh



import numpy as np

def geodesic_welding(a, b, s, t):
    with np.errstate(divide="ignore", invalid="ignore"):
        n = len(s)
        z = np.concatenate((s, a, [0]))
        w = np.concatenate((t, b, [np.inf]))
        z = step0(z, z[0], z[1])
        w = step0(w, w[0], w[1])
        for i in range(2, n):
            z = step1(z, z[i], 1j)
            w = step1(w, w[i], -1j)
        z = step2(z, z[0])
        w = step2(w, w[0])
        for i in range(n - 2, 0, -1):
            c1 = z[i]
            c2 = w[i]
            z = step3(z, c1, c2)
            w = step3(w, c1, c2)
        z = step4(z, z[0])
        w = step4(w, w[0])
        a = step5(z[n:-1], z[-1], w[-1])
        b = step5(w[n:-1], z[-1], w[-1])
        return a, b


def step0(z, p, q):
    w = np.sqrt((z - q) / (z - p))
    w[np.isinf(z)] = 1
    return w


def step1(z, p, m):
    c = np.real(p) / np.abs(p) ** 2
    d = np.imag(p) / np.abs(p) ** 2
    t = c * z / (1 + 1j * d * z)
    t[np.isinf(z)] = c / (1j * d)
    w = np.sqrt(t**2 - 1)
    k = np.imag(w) * np.imag(t) < 0
    w[k] = -w[k]
    w[z == 0] = m
    w[z == p] = 0
    return w


def step2(z, p):
    w = z / (1 - z / p)
    w[np.isinf(z)] = -p
    return w


def step3(z, p, q, eps=1e-10):
    p = np.imag(p)
    q = np.imag(q)
    p = p if p else eps
    q = q if q else eps
    a = -2 * p * q / (p - q)
    b = (p + q) / (p - q)
    r = z / (a - 1j * b * z)
    r[np.isnan(z)] = 1j / b
    r[np.isinf(z)] = 1j / b
    w = np.sqrt(r**2 + 1)
    s = np.imag(r) * np.imag(w) < 0
    w[s] = -w[s]
    return w


def step4(z, p):
    w = (z / (1 - z / p)) ** 2
    w[np.isinf(z)] = p**2
    w[np.isnan(w)] = np.inf
    return w


def step5(z, p, q):
    w = (z - p) / (z - q)
    w[np.isinf(z)] = 1
    return w



from setuptools import find_packages, setup

setup(
    name="hbs",
    version="0.3.0",
    author="Lin Chenran",
    author_email="chenranlin.17@gmail.com",
    description="Python library for computing Harmonic Beltrami Signature(HBS)",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ChanceAroundYou/hbs_python",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "opencv-python",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
